
import os
from collections import defaultdict
from itertools import combinations
import networkx as nx
import numpy as np
import gzip 
from Bio import SeqIO
import math
import csv
import cPickle as pkl
import  scipy.stats as stats
from goatools.obo_parser import GODag
import optparse
import warnings
warnings.filterwarnings('error')

class Tapas(object):
    def __init__(self,min_exp = 1, time_points = 30, obo_file = "gene_ontology.1_2.obo", ints_file= "ints_7227.pkl", string_file = "string_7227.pkl", homog_filt = 2.5, string_coff=800,corro_coff=0.7,ts_coff =0.5,goss_coff =6, gene_assoc_file ="gene_association.fb.gz", translation_file="dmel-all-translation-r5.45.fasta",tid_prof_file="tid_prof.pkl" ):
        
        self.min_exp = min_exp                  #Expression cut-off for a transcripts profile to be included   
        self.time_points = time_points          #Number of time points to use
        self.homog_filt= homog_filt             #GOSS homology filter for a cluster
        self.string_coff = string_coff          #cut off for STRING association inclusion
        self.corro_coff= corro_coff             #Minimum Pearson similarity cut off for two transcripts to be in the same cluster
        self.ts_coff = ts_coff                  #TS score cut off
        self.goss_coff= goss_coff               #GO semantic similarity cut off
        self.gene_assoc_file = gene_assoc_file  #File with GO annotations in the FlyBase annotation format
        self.translation_file =translation_file #Fasta file of protein sequences Flybase format
        self.tid_prof_file = tid_prof_file      #transcript expression profiles file contating a pickled dictionary where the transcript ID is the key and the expression profile is the value  
        self.ints_file=ints_file
        self.string_file =string_file
        self.obo_file =obo_file
        
        self.loadGo(annot_file=self.gene_assoc_file,obo_file=self.obo_file)
        self.setGoss()

        self.loadTranslation(self.translation_file)
        self.tid_prof =  pkl.load(open(os.path.join(self.tid_prof_file)))
        self.buildX()
        self.gid_ts_ordered={}
        self.pp=nx.read_gpickle(self.ints_file)
        self.ps=nx.read_gpickle(self.string_file)
        

        """get transcripts ordered by expression level"""
        
        x=defaultdict(list)
        for tid,prof in self.tid_prof.items():
            gid = self.tid_gid.get(tid)
            if gid:
                x[gid].append([np.mean(prof),tid])
        for gid, ts in x.items():
            ts.sort(reverse=True)
            self.gid_ts_ordered[gid]=[tid for m, tid in ts]
            
    def vcorrcoef(self,X,y):
        """Quick way of getting 1 vs all pearson correlations"""
        Xm = np.reshape(np.mean(X,axis=1),(X.shape[0],1))
        ym = np.mean(y)
        r_num = np.sum((X-Xm)*(y-ym),axis=1)
        r_den = np.sqrt(np.sum((X-Xm)**2,axis=1)*np.sum((y-ym)**2))
        r = r_num/r_den
        return r
    
    def buildX(self):
        print """Build a matrix of transcript profiles of size:"""
        
        X=[]    #to hold matrix of transcript profiles
        gids=[] #record corresponding gene ids
        oids=[] #record original transcript ids
        for tid,prof in self.tid_prof.items():
            
            gid = self.tid_gid.get(tid)
            if gid is None: continue
            prof=self.tid_prof[tid][0:self.time_points]
            if np.max(prof) <self.min_exp: continue
            X.append(prof)    
            gids.append(gid) 
            oids.append(tid)
        
        self.gids = np.array(gids)
        self.oids = np.array(oids)
        self.X=np.array(X)

        print self.X.shape
        
    
    
               
    def getTsScore(self,t1,t2): 
        nut=[self.tid_prof[t1][0:self.time_points],self.tid_prof[t2][0:self.time_points]]   
         
        U, s, V = np.linalg.svd(np.array(nut), full_matrices=False)
        
        s=s/np.sum(s)
        efs = 1-(s[0] -0.5)* 2
        return efs    


    def runQuery(self,query_gid):
       
 
        
        ts = self.gid_ts_ordered.get(query_gid,[])
        
        
        if len(ts)==0: return {"warning" : "No transcripts for the query gene ID, so will skip this gene"}
        if len(ts)==1: return {"warning" : "Only 1 transcript for the query gene ID, so will skip this gene"}
        tprim = ts[0]
        
        pps = []
        tprim =ts[0]
        for tsec in ts[1:]:
            
            pp=nx.Graph()
            ts_score= self.getTsScore(tprim, tsec)
            pp.ts_score = ts_score
            
            if self.tid_seq[tprim] == self.tid_seq[tsec]:
                continue 
        
            if ts_score < self.ts_coff:
                continue
            
            
            sinf=self.getClust(tsec,tother =tprim)
            
            
           
            
            for clust_gid,clust_tid,corro_query,corro_other in sinf.get("tclust"):    
                corro_dif = corro_query - corro_other
                if corro_dif < 0.5: continue#only keep things where theres a bick difference between primary and minor isoforms
                if clust_gid == query_gid: continue

                if self.ps.has_edge(query_gid,clust_gid):
                    
                    inf =self.ps[query_gid][clust_gid]   
                    if float(inf["weight"]) > self.string_coff:                 
                        pp.add_edge(clust_tid,tsec,{"string":inf["weight"]} )
                            
                
                if self.pp.has_edge(query_gid,clust_gid):
                    
                    inf =self.pp[query_gid][clust_gid]
                    pmids=set() 
                    for k, v in inf["db_ents"]: pmids |=v.get("pmids",set([]))
                    pp.add_edge(clust_tid,tsec,{"pmids":"|".join(list(map(str,pmids)))})
                    
                 
                goss,go=self.getGoss(self.eg_gos.get(query_gid,[]),self.eg_gos.get(clust_gid,[]))
                
                if goss > self.goss_coff:
                    #print goss,go
                    gacc=self.obo_dag.query_term(go,verbose=False)
                    if gacc:
                        pp.add_edge(clust_tid,tsec,{"term":gacc.name,"goss":goss})
                
                if pp.has_edge(clust_tid,tsec):
                    pp.add_edge(clust_tid,tsec,{"correlation_difference":corro_dif, "correlation":corro_query})       
            
            if True:
                sinf=self.getClust(tsec,tother =tprim,homog_check=True,limit=20,only_go_annot = True)    
                pp.homog = sinf.get("homog")        
                pp.clust = list(sinf.get("clust"))   
            
            pp.query_gid = query_gid 
            pp.query_tid = tsec
            pps.append(pp)    
          
        return {"networks":pps}            
         
            
            
    def getClust(self,qtid,tother = None,include_query=True,homog_check=False,limit=False,only_go_annot=False):
        
        
        inf = {}    
        gid = self.tid_gid[qtid]
        profo = self.tid_prof[qtid][0:self.time_points]         #query transcripts profile
        profo_other = self.tid_prof[tother][0:self.time_points]
        corr_pm = stats.pearsonr(profo,profo_other)[0]    
    
        
        ccs = self.vcorrcoef(self.X,profo)
        ccs_other = self.vcorrcoef(self.X,profo_other)


        x=np.argsort(ccs)[::-1]
        y=np.sort(ccs)[::-1] 
        yother=ccs_other[x] 
        
        clust=set()
        tclust=set()
        if include_query:
            clust.add(gid)
            tclust.add(tuple([gid,qtid,1.0,corr_pm]))
        
        tids=self.oids[x]     
        
        for counter,ceg in enumerate(self.gids[x]):
            corro = y[counter]
            corro2 = yother[counter]
            if ceg in clust: continue
            if (only_go_annot) and ceg not in self.eg_gos: continue
            clust.add(ceg)
            tclust.add(tuple([ceg,tids[counter],corro,corro2]))
            if limit and len(clust) >limit: break
            if (corro < self.corro_coff): break
            
    
        inf["clust"] = clust
        inf["tclust"] = tclust
        
        homog=None
        if homog_check:
            homog = self.getGossHomog(clust)   
        
        
         
        inf["homog"] = homog 
        
        if homog and (homog > self.homog_filt):inf["homog_pass"] = True 
        else:inf["homog_pass"] = False
        
       
        return inf
    
    
        
    def getGoss(self,agos,bgos):
        if len(agos)==0 or len(bgos)==0: return [0,""]
        x=[[self.gossd.get(go,0),go] for go in agos & bgos]
        x.sort(reverse=True)
        return x[0]


    def getSelfHomog(self,clust):
        y=[]
        for eg in clust:
            goss,term=self.getGoss(self.eg_gos[eg],self.eg_gos[eg])
            y.append(goss)
        return np.mean(y)
           
    def getGossHomog(self,clust):
        y=[]
        
        
        for comb in combinations(list(clust),2):
            if comb[0] not in self.eg_gos or comb[1] not in self.eg_gos: continue 
            goss,term=self.getGoss(self.eg_gos[comb[0]],self.eg_gos[comb[1]])
            y.append(goss)
        if len(y)==0: return 0
        return np.mean(y)
    
    
        

    def setGoss(self):
        freqs={}
        for go, egs in self.go_egs.items():
            freqs[go]=len(egs)
            
        x=[[freq,go] for go,freq in freqs.items()]
        x.sort(reverse=True)
        root_freq, root_node = x[0]
        root_freq=float(root_freq)
        self.gossd={}
        for go, freq in freqs.items():
            goss= -math.log( freq / root_freq)
            self.gossd[go]=goss 
            
            
    def loadGo(self, annot_file="", obo_file="", branches=["P"],efilts=["IEA","IEP"]): 
        print "loading gene ontology assignments"
        print os.path.join(obo_file)
        self.obo_dag = GODag(os.path.join(obo_file))
        
        eg_gos = defaultdict(set)
        go_egs = defaultdict(set)
        
        for line in gzip.open(os.path.join(annot_file)):
            if line.startswith("!"): continue
            vals=line.split("\t")
           
            if vals[6] in efilts: continue #for BP num
            if vals[6]=="ND": continue 
            if vals[3]=="NOT":
                continue
            acc =vals[4]
           

            gos = set([acc]) 
            gacc=self.obo_dag.query_term(acc,verbose=False)
            if gacc is None: continue
            if gacc.is_obsolete == True:continue
            
            branch=vals[8]
            eg = vals[1]
        
            if branch not in branches: continue
            
            paccs = gacc.get_all_parents()
            for pacc in paccs: gos.add(pacc)
            
            eg_gos[eg] |= gos
            for go in gos:go_egs[go].add(eg)
            
        self.go_egs = go_egs 
        self.eg_gos = eg_gos


    def loadTranslation(self,infile):
        print "loading protein coding sequences"
        
        self.tid_name = {}
        self.name_tid={}
        self.tid_gid = {}
        self.tid_seq={}
        self.gid_tids=defaultdict(set)
        for record in SeqIO.parse(open(os.path.join(infile)), "fasta"):
             
            trans=None
            parent=None
            name=None
            for i in record.description.split(";"):
                i=i.lstrip()
                if i.startswith("parent="):parent = i.split("=")[1]
                if i.startswith("name="):name = i.split("=")[1]
            if parent is None: 
                print "no parent"
                exit(1)

            
            gid,tid = parent.split(",")
            self.gid_tids[gid].add(tid)
            self.tid_gid[tid]=gid
            self.tid_name[tid]=name
            self.name_tid[name]=tid
            self.tid_seq[tid] = str(record.seq)

if __name__ == "__main__":

          
    parser = optparse.OptionParser()
    
    
    parser.add_option('--min-exp', action="store", 
                      default=1.0, 
                      dest="min_exp",
                      help="Expression cut-off for a transcripts profile to be included",
                      )
    parser.add_option('--time-points', action="store", 
                      default=30, 
                      dest="time_points",
                      help="Number of Expression profile time points to use",
                      )
    
    parser.add_option('--homog-filt', action="store", 
                      default=2.5, 
                      dest="homog_filt",
                      help="Average GOSS homology filter for a cluster to be considered functionally homogeneous",
                      )
    
    parser.add_option('--string-coff', action="store", 
                      default=800, 
                      dest="string_coff",
                      help="Cut off for STRING association inclusion",
                      )
    
    parser.add_option('--corr-coff', action="store", 
                      default=0.7, 
                      dest="corr_coff",
                      help="Minimum Pearson similarity cut off for two transcripts expression profiles for them to be in the same cluster",
                      )
    
    parser.add_option('--ts-coff', action="store", 
                      default=0.5, 
                      dest="ts_coff",
                      help="TS cut-off",
                      )
    
    parser.add_option('--goss-coff', action="store", 
                      default=6, 
                      dest="goss_coff",
                      help="GOSS cut-off for drawing links in the cytoscape output files",
                      )
    
    parser.add_option('--go-assoc', action="store", 
                      default="gene_association.fb.gz", 
                      dest="gene_assoc_file",
                      help="File with GO annotations in the (FlyBase annotation format)",
                      )
    
    
    parser.add_option('--fasta-file', action="store", 
                      default="dmel-all-translation-r5.45.fasta", 
                      dest="fasta_file",
                      help="Fasta file of protein sequences (Flybase format)",
                      )

    parser.add_option('--ints-file', action="store", 
                      default="ints_7227.pkl", 
                      dest="ints_file",
                      help="Graph of known protein interactions (networkx pickled format)",
                      )
    parser.add_option('--string-file', action="store", 
                      default="string_7227.pkl", 
                      dest="string_file",
                      help="Graph of string predicted interactions (networkx pickled format)",
                      )
    
    parser.add_option('--tid-prof', action="store", 
                      default="tid_prof.pkl", 
                      dest="tid_prof",
                      help="Expression profiles file contating a pickled dictionary where the transcript ID is the key and the expression profile is the value",
                      )

    parser.add_option('--out-dir', action="store", 
                      default="./cyto_hits", 
                      dest="odir",
                      help="Directory to write output results to",
                      )

    opts,args=parser.parse_args() 

    t = Tapas(min_exp = opts.min_exp, time_points = opts.time_points, homog_filt = opts.homog_filt, string_coff=opts.string_coff,corro_coff=opts.corr_coff,ts_coff =opts.ts_coff,goss_coff =opts.goss_coff, gene_assoc_file =opts.gene_assoc_file, translation_file=opts.fasta_file,tid_prof_file=opts.tid_prof )
    
    
    all_pps=[]
    for gid, ts in t.gid_tids.items():
        if len(ts)<2: continue #if we have less than two protein coding transcripts, skip
        #run query returns a network of the potential functional neighborood
        all_pps += t.runQuery(gid).get("networks",[])
    
    
    print """Write out the results to view in cytoscape"""
    bdir  = opts.odir
    
    
    ofh =open(os.path.join(bdir,"output.txt"),"w")
    sofh =open(os.path.join(bdir,"net.sif"),"w")
    
    nofh =open(os.path.join(bdir,"noats.txt"),"w")
    nofh.write("\t".join(["key","alias","score","type"]) + "\n")
    tid_score= {}
    
    
    eofh =open(os.path.join(bdir,"eats.txt"),"w")
    eofh.write("\t".join(["key","goss","term","pmids","ppi","stringo","corro","corro_diff"]) + "\n")
    
    ats=set()
    sif = nx.Graph()
    for pp in all_pps:

        ats |= set(pp.nodes())
        ofh.write("\t".join(["new query",  pp.query_gid, pp.query_tid]) + "\n")
        tid_score[pp.query_tid] = pp.ts_score
        
        for ea, eb,inf in pp.edges(data=True):
            ofh.write("\t".join(["information on functional links:"]) + "\n")
            for k,v in inf.items(): 
                ofh.write("\t".join([t.tid_name.get(ea,ea),t.tid_name.get(eb,eb), str(k),str(v)])+ "\n")  
            
            
            sif.add_edge(ea, eb)
            pmids = inf.get("pmids","")             
            ppi="no"
            if len(pmids) >0:ppi="yes"
            goss= str(inf.get("goss",""))
            term=str(inf.get("term",""))
            stringo =str(inf.get("string",""))
            corro=str(inf.get("correlation",0))
            corro_diff=str(inf.get("correlation_difference",0))
            x=[ea,eb]
            x.sort()
            eofh.write("\t".join([x[0] + " (pp) " + x[1],goss ,term, pmids, ppi,stringo,corro,corro_diff ])+ "\n" ) 
           
            
        if True:
            ofh.write("\t".join(["homology of the closest 20 neighbors:", str(pp.homog)]) + "\n")
            if pp.homog > opts.homog_filt:
                ofh.write("\t".join(["cluster:", ",".join(pp.clust)]) + "\n")    
        ofh.write("\n")
        
    for ea, eb in sif.edges(): 
        x=[ea,eb]
        x.sort()
        sofh.write(x[0] + " pp " + x[1] + "\n") 
        
                 
    for tid in ats:
        score = tid_score.get(tid,0)
        tsec = "yes" if  score > opts.ts_coff else "no"
        nofh.write("\t".join([tid,t.tid_name.get(tid),str(score),tsec])  + "\n") 
        
    
    
    
   
